﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using Classification.kNN;
using XmlClassifier.Algorithms;
using XmlClassifier.Algorithms.DistanceMeasures;
using XmlClassifier.Algorithms.DistanceMeasures.PTED;
using XmlClassifier.DataStructures.Patterns;
using XmlClassifier.Helpers;

namespace XmlClassifier.DataStructures
{
    [Serializable()]
	public class Document: IkNNable
	{
		/// <summary>
		/// Konstruktor
		/// </summary>
		/// <param name="doc">Dokument zapisany w formacie Zakiego. Pierwszy znak to klasa, do której należy.</param>
		public Document(string doc)
		{
			this.ExpectedClass = int.Parse(doc.Substring(0, doc.IndexOf(' ')));
			this.Content = doc.Substring(doc.IndexOf(' ') + 1, doc.Length - doc.IndexOf(' ') - 1);
			this.XDocument = this.Content.Parse(2);
		}

		/// <summary>
		/// Klasa, do której został przypisany dokument podczas fazy testowania.
		/// Dla dokumentów ze zbioru treningowego ta wartość powinna być zawsze pusta.
		/// </summary>
		public int? ActualClass { get; set; }

		/// <summary>
		/// Faktyczna klasa dokumentu.
		/// Wartość powinna być zawsze wypełniona.
		/// </summary>
		public int ExpectedClass { get; set; }

		/// <summary>
		/// Dokument zapisany w formacie Zakiego.
		/// </summary>
		public string Content { get; set; }

		/// <summary>
		/// Dokument zapisany w formacie XML
		/// </summary>    
        public XElement XDocument
        {
            get
            {
                return this.xDocument;
            }
            set
            {
                this.xDocument = value;
            }
        }

        [NonSerialized()]
        private XElement xDocument;

		#region kNN
		private double? distance;

		public void SetClass(string c)
		{
			this.ActualClass = int.Parse(c);
		}

		public void CalculateDistance(IClassifiable o)
		{
			if (o is Document)
			{
				this.distance = TED.Process(this.XDocument, ((Document)o).XDocument);
			}
			else if (o is TreePattern)
			{
				this.distance = PTED.Dynamic(((TreePattern)o).XPattern, this.XDocument);
			}
		}

		public double GetDistance()
		{
			if (this.distance.HasValue)
			{
				return this.distance.Value;
			}
			else
			{
				throw new NullReferenceException("The distance has not been calculated yet! Invoke the 'CalculateDistance' method first.");
			}
		}

		public double GetConfidence()
		{
			return 0;
		}

		public string GetClass()
		{
			return this.ExpectedClass.ToString();
		}

		public double GetWeight()
		{
			return 1;
		}
		#endregion
	}
}
