﻿using System.Collections.Generic;
using System.Linq;
using XmlClassifier.DataStructures;
using XmlClassifier.DataStructures.Patterns;

namespace XmlClassifier.DataStructures
{
	public class TreeMiner
	{
		public static List<TreePattern> ProcessDatabase(List<Document> documents, int minSup)
		{
			int[][] docs = documents.Select(x => x.Content.Split(' ').Select(y => int.Parse(y)).ToArray()).ToArray();

			TM.TreeMiner treeMiner = new TM.TreeMiner();

			unsafe
			{
				int** ptrs = stackalloc int*[docs.Length + 1];
				int numberOfDocuments = docs.Length;
				ptrs[0] = &numberOfDocuments;
				for (int i = 0; i < numberOfDocuments; i++)
				{
					fixed (int* ptr = docs[i])
					{
						ptrs[i + 1] = ptr;
					}
				}

				int** returned = treeMiner.Process(ptrs, minSup);

				Dictionary<string, int> result = new Dictionary<string, int>();
				int numberOfPatterns = returned[0][0];
				for (int i = 1; i < numberOfPatterns + 1; i++)
				{
					int numberOfElements = returned[i][0];
					string tmp = returned[i][1].ToString();
					for (int j = 2; j < numberOfElements; j++)
					{
						tmp += string.Format(" {0}", returned[i][j]);
					}
					result.Add(tmp, returned[i][numberOfElements]);
				}

				return result.Select(x => new TreePattern() { Support = x.Value / (double)documents.Count, Pattern = x.Key }).ToList();
			}
		}
	}
}
